from functools import reduce
from operator import or_

from django.core import checks
from django.db import models
from django.db.models import Q
from django.utils.text import slugify as django_slugify
from unidecode import unidecode


class AutoSlugField(models.SlugField):
    def __init__(self, populate_from, slugify=None, *args, **kwargs):
        self.populate_from = populate_from
        self.slugify = slugify or self._slugify

        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        kwargs['populate_from'] = self.populate_from

        if self.slugify != self._slugify:
            kwargs['slugify'] = self.slugify

        return name, path, args, kwargs

    def check(self, **kwargs):
        errors = []
        errors.extend(self._check_callable_attributes())
        errors.extend(self._check_self_reference())

        return errors

    def _check_callable_attributes(self):
        if self.slugify.__name__ == '<lambda>':
            return [
                checks.Error(
                    '`slugify` argument must be named top-level function.',
                    obj=self,
                    id='slugger.E001',
                )
            ]

        return []

    def _check_self_reference(self):
        if self.populate_from == self.attname:
            return [
                checks.Error(
                    '`populate_from` argument cannot reference its own '
                    'field name.',
                    obj=self,
                    id='slugger.E002',
                )
            ]

        return []

    def formfield(self, **kwargs):
        defaults = {
            'required': False,
        }

        defaults.update(kwargs)

        return super().formfield(**defaults)

    def pre_save(self, model_instance, add):
        value = super().pre_save(model_instance, add)

        if not value:
            from_field_value = getattr(model_instance, self.populate_from)
            value = self.slugify(from_field_value)[:self.max_length]

            if any((self.unique, self.unique_for_date,
                    self.unique_for_month, self.unique_for_year,
                    self.model._meta.unique_together)):
                value = self.get_unique_slug(value, model_instance)

            setattr(model_instance, self.attname, value)

        return value

    def _get_unique_lookups(self, instance):
        # Combine "uniqueness" lookups into single query to retrieve
        # objects which must have slugs different from this instance

        # slug must be globally unique, do not exclude anything
        if self.unique:
            return Q()

        lookups = []

        # `date` QuerySet filter lookup is not supported for DateField,
        # use separate lookups to handle both DateField and DateTimeField
        if self.unique_for_date:
            lookup_value = getattr(instance, self.unique_for_date)
            lookups.append({
                '%s__day' % self.unique_for_date: lookup_value.day,
                '%s__month' % self.unique_for_date: lookup_value.month,
                '%s__year' % self.unique_for_date: lookup_value.year,
            })

        if self.unique_for_month:
            lookup_value = getattr(instance, self.unique_for_month)
            lookups.append({
                '%s__month' % self.unique_for_month: lookup_value.month,
            })

        if self.unique_for_year:
            lookup_value = getattr(instance, self.unique_for_year)
            lookups.append({
                '%s__year' % self.unique_for_year: lookup_value.year,
            })

        def _get_unique_together_groups():
            for field_group in self.model._meta.unique_together:
                if self.attname in field_group:
                    yield (field_name for field_name in field_group
                           if field_name != self.attname)

        for field_group in _get_unique_together_groups():
            lookups.append({field_name: getattr(instance, field_name)
                            for field_name in field_group})

        return reduce(
            or_,
            (Q(**lookup) for lookup in lookups),
            Q()
        )

    def get_unique_slug(self, slug, instance):
        conflicts = self.model._default_manager.filter(
            ~Q(pk=instance.pk),
            self._get_unique_lookups(instance),
        )

        taken_slugs = sorted(conflicts.filter(
            **{'%s__regex' % self.attname: r'%s(-\d+)?$' % slug}
        ).values_list(self.attname, flat=True))

        if slug not in taken_slugs:
            return slug

        # generated slug is taken, remove it to ease searching for suffix
        taken_slugs.remove(slug)

        i = 1
        for value in taken_slugs:
            if not value.endswith(str(i)):
                break

            i += 1

        return '%s-%s' % (slug, i)

    @staticmethod
    def _slugify(value):
        return django_slugify(unidecode(value))



from django.db import models

from slugger.fields import AutoSlugField


class AutoSlugModel(models.Model):
    title = models.CharField(max_length=20)
    slug = AutoSlugField(populate_from='title', max_length=10)


class UniqueAutoSlugModel(models.Model):
    title = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from='title', unique=True)


class UniqueForAutoSlugModel(models.Model):
    title = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from='title', unique_for_date='unique_date',
                         unique_for_month='unique_month',
                         unique_for_year='unique_year')
    unique_date = models.DateField()
    unique_month = models.DateField()
    unique_year = models.DateField()


class UniqueForDateTimeAutoSlugModel(models.Model):
    title = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from='title', unique_for_date='pub_datetime')
    pub_datetime = models.DateTimeField()


class UniqueTogetherAutoSlugModel(models.Model):
    title = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from='title')
    field_1 = models.CharField(max_length=10)
    field_2 = models.CharField(max_length=10)
    field_3 = models.CharField(max_length=10)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = (
            ('created', 'field_1'),
            ('slug', 'field_1',),
            ('slug', 'field_2', 'field_3'),
        )


class ChildUniqueTogetherAutoSlugModel(UniqueTogetherAutoSlugModel):
    pass


class MixedUniqueAutoSlugModel(models.Model):
    title = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from='title', unique_for_date='unique_date')
    unique_date = models.DateField()
    field_1 = models.CharField(max_length=10)

    class Meta:
        unique_together = ('slug', 'field_1')


def custom_slugify(value):
    return 'custom-%s' % value


class CustomAutoSlugModel(models.Model):
    title = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from='title', slugify=custom_slugify)


class SelfReferenceAutoSlugModel(models.Model):
    slug = AutoSlugField(populate_from='slug')


class LambdaAutoSlugModel(models.Model):
    slug = AutoSlugField(populate_from='title', slugify=lambda: 'test')


class Child1Model(UniqueAutoSlugModel):
    pass


class Child2Model(UniqueAutoSlugModel):
    pass



from setuptools import setup, find_packages

setup(
    name='django-slugger',
    version='1.0.4',
    url='https://gitlab.com/dspechnikov/django-slugger/',
    description='Automatic slug field for Django.',
    license='MIT',

    author='Dmitry Pechnikov',
    author_email='dspechnikov@gmail.com',

    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=[
        'django',
        'unidecode',
    ],

    long_description=open('README.rst').read(),
    keywords='django field slug auto',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from datetime import date

import pytest
from django.core import checks

from slugger.fields import AutoSlugField
from tests.slugger.test_app import models


@pytest.mark.django_db()
class TestAutoSlugField:
    def test_generated_slug(self):
        obj = models.AutoSlugModel.objects.create(slug='', title='Test 1')

        assert obj.slug == 'test-1'

    def test_specified_slug(self):
        obj = models.UniqueAutoSlugModel.objects.create(
            slug='manual-slug',
            title='Test 1'
        )

        assert obj.slug == 'manual-slug', (
            'manual slug should override auto-generated one'
        )

    def test_exceeds_max_length(self):
        obj = models.AutoSlugModel.objects.create(title='x' * 20)

        assert obj.slug == 'x' * 10

    def test_add_suffix(self):
        models.UniqueAutoSlugModel.objects.create(slug='test')
        obj = models.UniqueAutoSlugModel.objects.create(slug='', title='Test')

        assert obj.slug == 'test-1'

    def test_generated_slug_match_slug_with_suffix(self):
        models.UniqueAutoSlugModel.objects.create(slug='test-1')
        obj = models.UniqueAutoSlugModel.objects.create(
            slug='', title='Test 1'
        )

        assert obj.slug == 'test-1-1', (
            'add suffix instead of incrementing existing suffix'
        )

    def test_suffix_increment(self):
        models.UniqueAutoSlugModel.objects.bulk_create([
            models.UniqueAutoSlugModel(slug='test'),
            models.UniqueAutoSlugModel(slug='test-1'),
            models.UniqueAutoSlugModel(slug='test-3'),
        ])
        obj = models.UniqueAutoSlugModel.objects.create(slug='', title='Test')

        assert obj.slug == 'test-2', (
            'suffix increment should be the smallest available value'
        )

    def test_regenerate_slug(self):
        obj = models.UniqueAutoSlugModel.objects.create(slug='', title='Test')
        obj.slug = ''
        obj.save()

        assert obj.slug == 'test', (
            'object slug should not be used for suffix generation'
        )

    def test_only_suffixed_slug_exists(self):
        models.UniqueAutoSlugModel.objects.create(slug='test-1')
        obj = models.UniqueAutoSlugModel.objects.create(slug='', title='Test')

        assert obj.slug == 'test', (
            'do not use suffix if original slug is not taken'
        )

    def test_unique_for_date(self):
        models.UniqueForAutoSlugModel.objects.create(
            slug='test', unique_date=date(2000, 1, 1),
            unique_month=date(2000, 1, 1), unique_year=date(2000, 1, 1)
        )

        unique_for_date = models.UniqueForAutoSlugModel.objects.create(
            slug='', title='test', unique_date=date(2000, 1, 1),
            unique_month=date(2000, 2, 1), unique_year=date(2001, 1, 1)
        )

        assert unique_for_date.slug == 'test-1'

    def test_unique_for_month(self):
        models.UniqueForAutoSlugModel.objects.create(
            slug='test', unique_date=date(2000, 1, 1),
            unique_month=date(2000, 1, 1), unique_year=date(2000, 1, 1)
        )

        unique_for_month = models.UniqueForAutoSlugModel.objects.create(
            slug='', title='test', unique_date=date(2000, 1, 2),
            unique_month=date(2000, 1, 1), unique_year=date(2001, 1, 1)
        )

        assert unique_for_month.slug == 'test-1'

    def test_unique_for_year(self):
        models.UniqueForAutoSlugModel.objects.create(
            slug='test', unique_date=date(2000, 1, 1),
            unique_month=date(2000, 1, 1), unique_year=date(2000, 1, 1)
        )

        unique_for_year = models.UniqueForAutoSlugModel.objects.create(
            slug='', title='test', unique_date=date(2000, 1, 2),
            unique_month=date(2000, 2, 1), unique_year=date(2000, 1, 1)
        )

        assert unique_for_year.slug == 'test-1'

    def test_multiple_unique_together_constraints(self):
        # unique_together = (
        #     ('created', 'field_1'),
        #     ('slug', 'field_1',),
        #     ('slug', 'field_2', 'field_3'),
        # )
        models.UniqueTogetherAutoSlugModel.objects.create(
            slug='test', field_1='1-1', field_2='2-1', field_3='3-1',
        )

        # violate only field_1 constraint
        field_1 = models.UniqueTogetherAutoSlugModel.objects.create(
            slug='', title='test', field_1='1-1', field_2='2-2', field_3='3-2',
        )
        # violate only field_2/field_3 constraint
        field_2_3 = models.UniqueTogetherAutoSlugModel.objects.create(
            slug='', title='test', field_1='1-2', field_2='2-1', field_3='3-1',
        )
        # violate both constraints
        both = models.UniqueTogetherAutoSlugModel.objects.create(
            slug='', title='test', field_1='1-1', field_2='2-2', field_3='3-2',
        )

        assert field_1.slug == 'test-1'
        assert field_2_3.slug == 'test-1'
        assert both.slug == 'test-2'

    def test_unique_together_multi_table_inheritance(self):
        models.ChildUniqueTogetherAutoSlugModel.objects.create(
            slug='test', field_1='1-1', field_2='2-1', field_3='3-1',
        )

        obj = models.ChildUniqueTogetherAutoSlugModel.objects.create(
            title='test', slug='', field_1='1-2', field_2='2-1', field_3='3-1',
        )

        assert obj.slug == 'test-1'

    def test_both_unique_for_and_unique_together(self):
        # unique_together = ('slug', 'field_1')
        models.MixedUniqueAutoSlugModel.objects.create(
            slug='test', unique_date=date(2000, 1, 1), field_1='1-1',
        )

        unique_for = models.MixedUniqueAutoSlugModel.objects.create(
            slug='', title='test', unique_date=date(2000, 1, 1), field_1='1-2',
        )
        unique_together = models.MixedUniqueAutoSlugModel.objects.create(
            slug='', title='test', unique_date=date(2000, 1, 2), field_1='1-1',
        )

        assert unique_for.slug == 'test-1'
        assert unique_together.slug == 'test-1'

    def test_custom_slugify_function(self):
        obj = models.CustomAutoSlugModel.objects.create(slug='', title='Test')

        assert obj.slug == 'custom-Test'

    def test_callable_attribute_check(self):
        field = next(f for f in models.LambdaAutoSlugModel._meta.get_fields()
                     if f.attname == 'slug')

        errors = [(error.level, error.id) for error in field.check()]

        assert (checks.ERROR, 'slugger.E001') in errors

    def test_self_reference_check(self):
        field = next(
            f for f in models.SelfReferenceAutoSlugModel._meta.get_fields()
            if f.attname == 'slug'
        )

        errors = [(error.level, error.id) for error in field.check()]

        assert (checks.ERROR, 'slugger.E002') in errors

    def test_form_field_is_not_required(self):
        field = AutoSlugField(populate_from='title')

        assert not field.formfield().required

    def test_field_deconstruction(self):
        field = AutoSlugField(
            populate_from='title', slugify=models.custom_slugify
        )

        name, path, args, kwargs = field.deconstruct()

        assert kwargs['populate_from'] == 'title'
        assert kwargs['slugify'] == models.custom_slugify

    def test_model_inheritance(self):
        models.Child1Model.objects.create(title='test', slug='')

        obj = models.Child2Model.objects.create(title='test', slug='')

        assert obj.slug == 'test-1'
